# date parts as integers from date in iso-8601
year = "toInt16OrZero(substring({date_string}, 1, 4))"
month = "toInt16OrZero(substring({date_string}, 6, 2))"
day = "toInt16OrZero(substring({date_string}, 9, 2))"

one_if_leap_year_else_zero = f"""
CASE
    WHEN ({year} % 4 = 0 AND {year} % 100 <> 0) OR {year} % 400 = 0 THEN 1
    ELSE 0
END
"""

# hard-code some results to cut re-computations
# intDiv(1970, 4) == 492
FLOOR_1970_OVER_4 = 492
# intDiv(1970, 100) == 19
FLOOR_1970_OVER_100 = 19
# intDiv(1970, 400) == 4
FLOOR_1970_OVER_400 = 4

# number of leap years between given year and 1970
# not counting the year itself
leap_years_since_1970 = f"""
    (intDiv({year} - 1, 4) - {FLOOR_1970_OVER_4})
    - (intDiv({year} - 1, 100) - {FLOOR_1970_OVER_100})
    + (intDiv({year} - 1, 400) - {FLOOR_1970_OVER_400})
"""
days_from_epoch_to_start_of_year = f"""
({year} - 1970) * 365 + {leap_years_since_1970}
"""

# cumulative days in non-leap years:
# J: 31
# F: 28 + 31 = 59
# M: 31 + 59 = 90
# A: 30 + 90 = 120
# M: 31 + 120 = 151
# J: 30 + 151 = 181
# J: 31 + 181 = 212
# A: 31 + 212 = 243
# S: 30 + 243 = 273
# O: 31 + 273 = 304
# N: 30 + 304 = 334
# D: 31 + 334 = 365 (which we don't need, but is a good check)
# it's to start of month, so we need cumulative total to month before
days_from_start_of_year_to_start_of_month = f"""
CASE
    WHEN {month} = 1 THEN 0
    WHEN {month} = 2 THEN 31
    WHEN {month} = 3 THEN 59 + {one_if_leap_year_else_zero}
    WHEN {month} = 4 THEN 90 + {one_if_leap_year_else_zero}
    WHEN {month} = 5 THEN 120 + {one_if_leap_year_else_zero}
    WHEN {month} = 6 THEN 151 + {one_if_leap_year_else_zero}
    WHEN {month} = 7 THEN 181 + {one_if_leap_year_else_zero}
    WHEN {month} = 8 THEN 212 + {one_if_leap_year_else_zero}
    WHEN {month} = 9 THEN 243 + {one_if_leap_year_else_zero}
    WHEN {month} = 10 THEN 273 + {one_if_leap_year_else_zero}
    WHEN {month} = 11 THEN 304 + {one_if_leap_year_else_zero}
    WHEN {month} = 12 THEN 334 + {one_if_leap_year_else_zero}
END
"""
days_since_start_of_month = f"{day} - 1"

days_since_epoch_template = f"""
{days_from_epoch_to_start_of_year}
    + {days_from_start_of_year_to_start_of_month}
    + {days_since_start_of_month}
"""

days_since_epoch_sql = days_since_epoch_template.format(date_string="date_string")



import pandas as pd
from clickhouse_connect.driver.client import Client

from ..database_api import ClickhouseAPI
from .dataframe import ClickhouseServerDataFrame


class ClickhouseServerAPI(ClickhouseAPI):
    def __init__(
        self,
        client: Client,
        register_custom_udfs: bool = True,
    ):
        super().__init__()

        self.client = client
        self.set_union_default_mode()
        self._create_random_function()
        if register_custom_udfs:
            self._register_custom_udfs()

    def _table_registration(self, input, table_name) -> None:
        input = self._coerce_input_to_pd_if_needed(input)

        if isinstance(input, pd.DataFrame):
            sql = self._create_table_sql_from_pd_frame(input, table_name)
            self._execute_sql_against_backend(sql)
            self.client.insert_df(table_name, input)
        else:
            raise TypeError(
                "ClickhouseServerAPI currently only accepts table names (str) "
                "or pandas DataFrames as inputs for table registration. "
                f"Received type {type(input)}"
            )

    def table_to_splink_dataframe(self, templated_name, physical_name):
        return ClickhouseServerDataFrame(templated_name, physical_name, self)

    def table_exists_in_database(self, table_name):
        sql = self._information_schema_query(
            "table_name", "tables", table_name, self.database
        )

        res = self.client.query(sql).result_set
        return len(res) > 0

    def _execute_sql_against_backend(
        self, final_sql: str, templated_name: str = None, physical_name: str = None
    ):
        self.client.query(final_sql)

    def _get_results_from_backend(self, sql: str):
        res = self.client.query(sql).named_results()
        return res

    @property
    def database(self) -> str:
        return self.client.database or "default"

    # Clickhouse can not handle a bare 'UNION' by default
    # we can set desired behaviour for the session by executing this
    def set_union_default_mode(self) -> None:
        """
        Set `union_default_mode` to 'DISTINCT'

        Splink uses SQL involving bare 'UNION's in order to combine rows _without_
        removing duplicates.
        Clickhouse requires this to be explicitly set, so this function changes the
        server setting so that this is the default behaviour.
        See [the Clickhouse docs](https://clickhouse.com/docs/en/sql-reference/statements/select/union).

        This happens automatically on creation of `ClickhouseServerAPI`.
        However, this setting is tied to a session, so if this expires then this
        configuration value will be forgotten, and you may receive errors such as
        `Expected ALL or DISTINCT in SelectWithUnion query`.
        In this case you will need to manually call this function.
        """  # noqa: E501 (line too long)
        self._execute_sql_against_backend("SET union_default_mode = 'DISTINCT'")

    def _create_table_sql_from_pd_frame(self, df: pd.DataFrame, table_name: str) -> str:
        sql = f"CREATE OR REPLACE TABLE {table_name} ("

        first_col = True
        for column_name in df.columns:
            if not first_col:
                sql += ", "
            column = df[column_name]
            col_type = column.dtype
            first_col = False

            if pd.api.types.is_unsigned_integer_dtype(col_type):
                sql += f"{column_name} Nullable(UInt32)"
            elif pd.api.types.is_integer_dtype(col_type):
                sql += f"{column_name} Nullable(Int32)"
            elif pd.api.types.is_float_dtype(col_type):
                sql += f"{column_name} Nullable(Float64)"
            elif pd.api.types.is_list_like(column[0]):
                sql += f"{column_name} Array(String)"
            elif pd.api.types.is_string_dtype(col_type):
                sql += f"{column_name} Nullable(String)"
            elif pd.api.types.is_datetime64_dtype(col_type):
                sql += f"{column_name} Nullable(DateTime64)"
            else:
                raise ValueError(f"Unknown data type {col_type}")

        sql += ") ENGINE MergeTree ORDER BY tuple()"

        return sql



from abc import abstractmethod
from typing import Optional

import pandas as pd
from splink.internals.database_api import DatabaseAPI

from .custom_sql import days_since_epoch_sql
from .dialect import ClickhouseDialect


class ClickhouseAPI(DatabaseAPI[None]):
    sql_dialect = ClickhouseDialect()

    @property
    def database(self) -> Optional[str]:
        return None

    @abstractmethod
    def _get_results_from_backend(self, sql: str):
        pass

    def _coerce_input_to_pd_if_needed(self, input):
        if isinstance(input, dict):
            input = pd.DataFrame(input)
        elif isinstance(input, list):
            input = pd.DataFrame.from_records(input)
        return input

    def _information_schema_query(
        self,
        column_name: str,
        information_schema_table_name: str,
        table_name: str,
        schema_name: Optional[str] = None,
    ):
        and_if_needed = (
            f"AND table_schema = '{schema_name}'" if schema_name is not None else ""
        )
        return f"""
            SELECT {column_name}
            FROM information_schema.{information_schema_table_name}
            WHERE table_name = '{table_name}'
            {and_if_needed}
        """

    @property
    def _core_replacements(self) -> list[tuple[str, str]]:
        return [
            ("float8", "Float64"),
            # workaround for https://github.com/ClickHouse/ClickHouse/issues/61004
            ("count(*)", "count()"),
            ("COUNT(*)", "COUNT()"),
            # TODO: very sorry for this
            # avoids 'double selection' issue in creating __splink__block_counts
            (", count_l, count_r,", ","),
            (
                "SELECT DISTINCT r.representative",
                "SELECT DISTINCT r.representative AS representative",
            ),
        ]

    @property
    def _specific_replacements(self) -> list[tuple[str, str]]:
        return []

    # some excessively brittle SQL replacements to handle Clickhouse name-resolution
    # and other dialect quirks
    @property
    def _sql_replacements(self) -> list[tuple[str, str]]:
        return self._core_replacements + self._specific_replacements

    def _setup_for_execute_sql(self, sql: str, physical_name: str) -> str:
        self.delete_table_from_database(physical_name)
        for old_sql, new_sql in self._sql_replacements:
            sql = sql.replace(old_sql, new_sql)

        sql = f"CREATE TABLE {physical_name} ORDER BY tuple() AS {sql}"
        return sql

    # alias random -> rand. Need this function for comparison viewer
    def _create_random_function(self) -> None:
        self._execute_sql_against_backend(
            "CREATE FUNCTION IF NOT EXISTS random AS () -> rand()"
        )

    def _register_custom_udfs(self) -> None:
        self._execute_sql_against_backend(
            f"""
            CREATE FUNCTION IF NOT EXISTS
                days_since_epoch AS
                (date_string) -> {days_since_epoch_sql}
            """
        )



from typing import Any

from .clickhouse_server.database_api import ClickhouseServerAPI

__version__ = "0.4.1"


# Use getarr to make the error appear at the point of use
def __getattr__(name: str) -> Any:
    try:
        if name == "ChDBAPI":
            from .chdb.database_api import ChDBAPI

            return ChDBAPI
    except ImportError as err:
        if name == "ChDBAPI":
            raise ImportError(
                f"'{name}' cannot be imported because its dependencies are not "
                "installed. To get these please install with the 'chdb' extras: "
                "`pip install splinkclickhouse[chdb]`."
            ) from err
    raise ImportError(f"cannot import name '{name}' from splinkclickhouse") from None


__all__ = ["ChDBAPI", "ClickhouseServerAPI"]



from copy import copy

from splink.internals.column_expression import (
    ColumnExpression as SplinkColumnExpression,
)
from splink.internals.dialects import SplinkDialect

from .dialect import ClickhouseDialect


class ColumnExpression(SplinkColumnExpression):
    def __init__(self, sql_expression: str):
        super().__init__(sql_expression=sql_expression, sql_dialect=ClickhouseDialect())

    def _parse_date_to_int_dialected(
        self, name: str, sql_dialect: SplinkDialect
    ) -> str:
        # need to have sql_dialect passed, even if it does nothing
        # parent requires this to be in function signature

        return f"days_since_epoch({name})"

    def parse_date_to_int(self) -> "ColumnExpression":
        """
        Parses date string to an integer, representing days since
        the Unix epoch (1970-01-01).
        """
        clone = self._clone()
        clone.operations.append(clone._parse_date_to_int_dialected)
        return clone

    @staticmethod
    def from_base_expression(
        column_expression: SplinkColumnExpression,
    ) -> "ColumnExpression":
        new_expression = ColumnExpression(
            sql_expression=column_expression.raw_sql_expression
        )
        new_expression.operations = copy(column_expression.operations)
        return new_expression



import chdb.dbapi as chdb_dbapi

from ..database_api import ClickhouseAPI
from .dataframe import ChDBDataFrame


class ChDBAPI(ClickhouseAPI):
    def __init__(
        self,
        con: chdb_dbapi.Connection,
        schema: str = "splink",
        register_custom_udfs: bool = True,
    ):
        super().__init__()

        self.con = con
        self._db_schema = schema

        self._create_splink_schema()
        self._create_random_function()
        if register_custom_udfs:
            self._register_custom_udfs()

    def _table_registration(self, input, table_name):
        input = self._coerce_input_to_pd_if_needed(input)

        # chdb currently needs pandas indices to start at 0
        # see https://github.com/chdb-io/chdb/issues/282
        # reset the index if not the case, but otherwise leave alone
        # TODO: remove this workaround once chdb issue is resolved
        try:
            input[0]
        except KeyError:
            input = input.reset_index()
        sql = (
            f"CREATE OR REPLACE TABLE {self._db_schema}.{table_name} "
            "ORDER BY tuple() "
            f"AS SELECT * FROM Python(input);"
        )
        self._execute_sql_against_backend(sql)

    def table_to_splink_dataframe(self, templated_name, physical_name):
        return ChDBDataFrame(templated_name, physical_name, self)

    def table_exists_in_database(self, table_name):
        sql = self._information_schema_query(
            "table_name", "tables", table_name, self.database
        )

        cursor = self._get_cursor()
        try:
            cursor.execute(sql)
            table_name_row = cursor.fetchone()
        finally:
            self._reset_cursor(cursor)
        return table_name_row is not None

    @property
    def _specific_replacements(self) -> list[tuple[str, str]]:
        return [
            # TODO: horrible hack
            # can't seem to set union_default_mode for some reason
            ("UNION ALL", "__tmp__ua__"),
            ("UNION", "UNION DISTINCT"),
            ("__tmp__ua__", "UNION ALL"),
        ]

    def _execute_sql_against_backend(
        self, final_sql: str, templated_name: str = None, physical_name: str = None
    ) -> None:
        cursor = self._get_cursor()
        try:
            cursor.execute(final_sql)
        finally:
            self._reset_cursor(cursor)
        return None

    def _get_results_from_backend(self, sql: str):
        cursor = self._get_cursor()
        try:
            cursor.execute(sql)
            res = cursor.fetchall()
        finally:
            self._reset_cursor(cursor)
        return res

    def _get_cursor(self) -> chdb_dbapi.cursors.DictCursor:
        return self.con.cursor(chdb_dbapi.cursors.DictCursor)

    def _reset_cursor(self, cursor):
        cursor.close()

    def _create_splink_schema(self):
        sql = f"""
        CREATE DATABASE IF NOT EXISTS {self._db_schema};
        USE {self._db_schema};
        """
        self._execute_sql_against_backend(sql)



# need this for annotating db_api:
from __future__ import annotations

from typing import TYPE_CHECKING

from ..dataframe import ClickhouseDataFrame

if TYPE_CHECKING:  # pragma: no cover
    from .database_api import ClickhouseServerAPI


class ClickhouseServerDataFrame(ClickhouseDataFrame):
    db_api: ClickhouseServerAPI



from __future__ import annotations

from typing import Iterable, Literal

import splink.comparison_level_library as cll
from splink.internals.column_expression import ColumnExpression
from splink.internals.comparison_creator import ComparisonCreator
from splink.internals.comparison_level_creator import ComparisonLevelCreator
from splink.internals.comparison_library import (
    AbsoluteTimeDifferenceAtThresholds as SplinkAbsoluteTimeDifferenceAtThresholds,
)
from splink.internals.comparison_library import (
    DateMetricType,
    _DamerauLevenshteinIfSupportedElseLevenshteinLevel,
)
from splink.internals.comparison_library import (
    DateOfBirthComparison as SplinkDateOfBirthComparison,
)
from splink.internals.comparison_library import (
    PairwiseStringDistanceFunctionAtThresholds as SplinkPairwiseStringDistanceFunctionAtThresholds,  # noqa: E501 (can't keep format and check happy)
)
from splink.internals.misc import ensure_is_iterable

import splinkclickhouse.comparison_level_library as cll_ch

from .column_expression import ColumnExpression as CHColumnExpression


class DistanceInKMAtThresholds(ComparisonCreator):
    def __init__(
        self,
        lat_col: str,
        long_col: str,
        km_thresholds: Iterable[float] | float,
        calculation_method: Literal["great_circle", "wgs84"] = "wgs84",
    ):
        """
        A comparison of the latitude, longitude coordinates defined in
        'lat_col' and 'long col' giving the distance between them in km.

        An example of the output with km_thresholds = [1, 10] would be:

        - The two coordinates are within 1 km of one another
        - The two coordinates are within 10 km of one another
        - Anything else (i.e. the distance between coordinates are > 10km apart)

        This uses in-built Clickhouse functions to calculate the distance,
        either based on a spherical great-circle model, or based on WGS 84.

        Args:
            lat_col (str): The name of the latitude column to compare.
            long_col (str): The name of the longitude column to compare.
            km_thresholds (iterable[float] | float): The km threshold(s) for the
                distance levels.
            calculation_method (str): The method to use to compute distances from
                latitude and longitude.
                Can be 'great_circle' for the great circle distance, based on a
                spherical model (Clickhouse `greatCircleDistance`), or 'wgs84'
                to use the more accurate WGS 84 Ellipsoid model
                (Clickhouse `geoDistance`). Default 'wgs84'.
        """

        thresholds_as_iterable = ensure_is_iterable(km_thresholds)
        self.thresholds = [*thresholds_as_iterable]
        self.calculation_method = calculation_method
        super().__init__(
            col_name_or_names={
                "latitude_column": lat_col,
                "longitude_column": long_col,
            }
        )

    def create_comparison_levels(self) -> list[ComparisonLevelCreator]:
        lat_col = self.col_expressions["latitude_column"]
        long_col = self.col_expressions["longitude_column"]
        return [
            cll.Or(cll.NullLevel(lat_col), cll.NullLevel(long_col)),
            *[
                cll_ch.DistanceInKMLevel(
                    lat_col,
                    long_col,
                    km_threshold=threshold,
                    calculation_method=self.calculation_method,
                )
                for threshold in self.thresholds
            ],
            cll.ElseLevel(),
        ]

    def create_output_column_name(self) -> str:
        lat_col = self.col_expressions["latitude_column"]
        long_col = self.col_expressions["longitude_column"]
        return f"{lat_col.output_column_name}_{long_col.output_column_name}"


class ExactMatchAtSubstringSizes(ComparisonCreator):
    def __init__(
        self,
        col_name: str,
        substring_size_or_sizes: Iterable[int] | int = [4, 3, 2],
        include_full_exact_match: bool = True,
    ):
        """
        A comparison between columns at several sizes of substring

        An example of the output with substring_size_or_sizes = [3, 1] and
        include_full_exact_match = True would be:

        - The two columns match exactly
        - The substring of the first three characters of each column match exactly
        - The substring of the first character of each column match exactly

        This is suitable for a hierarchically structured string, such as a geohash
        column (from e.g. the result of `geohashEncode`).
        See https://clickhouse.com/docs/en/sql-reference/functions/geo/geohash.

        Args:
            col_name (str): The name of the column to compare.
            substring_size_or_sizes (iterable[int] | int): The size(s) of the substrings
                to compare, taken from the start of the string.
                Default [4, 3, 2]
            include_full_exact_match (bool): Whether or not to include a level for
                an exact match on the full string.
                Defaults to True.
        """

        sizes_as_iterable = ensure_is_iterable(substring_size_or_sizes)
        self.substring_sizes = [*sizes_as_iterable]
        self.include_full_exact_match = include_full_exact_match
        super().__init__(col_name)

    def create_comparison_levels(self) -> list[ComparisonLevelCreator]:
        raw_column = self.col_expression

        levels = [cll.NullLevel(raw_column)]
        if self.include_full_exact_match:
            levels.append(cll.ExactMatchLevel(raw_column))
        levels.extend(
            cll.ExactMatchLevel(raw_column.substr(1, threshold))
            for threshold in self.substring_sizes
        )
        levels.append(cll.ElseLevel())
        return levels

    def create_output_column_name(self) -> str:
        return self.col_expression.output_column_name


class AbsoluteDateDifferenceAtThresholds(SplinkAbsoluteTimeDifferenceAtThresholds):
    def __init__(
        self,
        col_name: str,
        *,
        input_is_string: bool,
        metrics: DateMetricType | list[DateMetricType],
        thresholds: float | list[float],
        term_frequency_adjustments: bool = False,
        invalid_dates_as_null: bool = True,
    ):
        """
        Represents a comparison of the data in `col_name` with multiple levels based on
        absolute time differences. For more details see Splink docs.

        In database this represents data as an integer counting number of days since
        1970-01-01 (Unix epoch).
        The input data can be either a string in YYYY-MM-DD format, or an
        integer of the number days since the epoch.

        Args:
            col_name (str): The name of the column to compare.
            input_is_string (bool): If True, the input dates are treated as strings
                and parsed to integers, and must be in ISO 8601 format.
            metrics (Union[DateMetricType, List[DateMetricType]]): The unit(s) of time
                to use when comparing dates. Can be 'second', 'minute', 'hour', 'day',
                'month', or 'year'.
            thresholds (Union[int, float, List[Union[int, float]]]): The threshold(s)
                to use for the time difference level(s).
            term_frequency_adjustments (bool, optional): Whether to apply term frequency
                adjustments. Defaults to False.
            invalid_dates_as_null (bool, optional): If True and `input_is_string` is
                True, treat invalid dates as null. Defaults to True.
        """
        super().__init__(
            col_name,
            input_is_string=input_is_string,
            metrics=metrics,
            thresholds=thresholds,
            datetime_format=None,
            term_frequency_adjustments=term_frequency_adjustments,
            invalid_dates_as_null=invalid_dates_as_null,
        )

    @property
    def datetime_parse_function(self):
        return lambda fmt: CHColumnExpression.from_base_expression(
            self.col_expression
        ).parse_date_to_int()

    @property
    def cll_class(self):
        # the parent class will try to pass datetime_format argument
        # our child class doesn't accept that, so we just ditch it
        def level_factory(*args, datetime_format=None, **kwargs):
            return cll_ch.AbsoluteDateDifferenceLevel(*args, **kwargs)

        return level_factory


class DateOfBirthComparison(SplinkDateOfBirthComparison):
    def __init__(
        self,
        col_name: str | ColumnExpression,
        *,
        input_is_string: bool,
        datetime_thresholds: float | Iterable[float] = (1, 1, 10),
        datetime_metrics: DateMetricType | Iterable[DateMetricType] = (
            "month",
            "year",
            "year",
        ),
        invalid_dates_as_null: bool = True,
    ):
        """
        Generate an 'out of the box' comparison for a date of birth column
        in the `col_name` provided. For more details see Splink docs.

        In database this represents data as an integer counting number of days since
        1970-01-01 (Unix epoch).
        The input data can be either a string in YYYY-MM-DD format, or an
        integer of the number days since the epoch.

        Args:
            col_name (str): The name of the column to compare.
            input_is_string (bool): If True, the input dates are treated as strings
                and parsed to integers, and must be in ISO 8601 format.
            datetime_thresholds (Union[int, float, List[Union[int, float]]], optional):
                Numeric thresholds for date differences. Defaults to [1, 1, 10].
            datetime_metrics (Union[DateMetricType, List[DateMetricType]], optional):
                Metrics for date differences. Defaults to ["month", "year", "year"].
            invalid_dates_as_null (bool, optional): If True, treat invalid dates as null
                as opposed to allowing e.g. an exact or levenshtein match where one side
                or both are an invalid date.  Only used if input is a string.  Defaults
                to True.
        """
        super().__init__(
            col_name=col_name,
            input_is_string=input_is_string,
            datetime_thresholds=datetime_thresholds,
            datetime_metrics=datetime_metrics,
            datetime_format=None,
        )

    @property
    def datetime_parse_function(self):
        return lambda fmt: CHColumnExpression.from_base_expression(
            self.col_expression
        ).parse_date_to_int()

    def create_comparison_levels(self) -> list[ComparisonLevelCreator]:
        # pretty much a copy of the Splink version, but unlike
        # AbsoluteDateDifferenceAtThresholds this does not allow a way
        # to hook in a different date-difference comparison level
        # so we duplicate, and just switch the date level
        if self.invalid_dates_as_null and self.input_is_string:
            null_col = self.datetime_parse_function(self.datetime_format)
        else:
            null_col = self.col_expression

        levels: list[ComparisonLevelCreator] = [
            cll.NullLevel(null_col),
        ]

        levels.append(
            cll.ExactMatchLevel(self.col_expression).configure(
                label_for_charts="Exact match on date of birth"
            )
        )

        if self.input_is_string:
            col_expr_as_string = self.col_expression
        else:
            col_expr_as_string = self.col_expression.cast_to_string()

        levels.append(
            _DamerauLevenshteinIfSupportedElseLevenshteinLevel(
                col_expr_as_string, distance_threshold=1
            ).configure(label_for_charts="DamerauLevenshtein distance <= 1")
        )

        # we always have at least one threshold+metric, fixed in super().__init__
        for threshold, metric in zip(self.datetime_thresholds, self.datetime_metrics):
            levels.append(
                cll_ch.AbsoluteDateDifferenceLevel(
                    self.col_expression,
                    threshold=threshold,
                    metric=metric,
                    input_is_string=self.input_is_string,
                ).configure(
                    label_for_charts=f"Abs date difference <= {threshold} {metric}"
                )
            )

        levels.append(cll.ElseLevel())
        return levels


class PairwiseStringDistanceFunctionAtThresholds(
    SplinkPairwiseStringDistanceFunctionAtThresholds
):
    def create_comparison_levels(self) -> list[ComparisonLevelCreator]:
        return [
            cll.NullLevel(self.col_expression),
            # It is assumed that any string distance treats identical
            # arrays as the most similar
            cll.ArrayIntersectLevel(self.col_expression, min_intersection=1),
            *[
                cll_ch.PairwiseStringDistanceFunctionLevel(
                    self.col_expression,
                    distance_threshold=threshold,
                    distance_function_name=self.distance_function_name,
                )
                for threshold in self.thresholds
            ],
            cll.ElseLevel(),
        ]
